#!/usr/bin/env python

from docopt import docopt
from .server import start


cmd = """Python LiveReload

Usage:
    livereload [-p <port>|--port=<port>] [-b|--browser] [<directory>]

Options:
    -h --help                       show this screen
    -p <port> --port=<port>         specify a server port, default is 35729
    -b --browser                    open browser when start server
"""


def main():
    args = docopt(cmd)
    port = args.get('--port')
    root = args.get('<directory>')
    autoraise = args.get('--browser')
    if port:
        port = int(port)
    else:
        port = 35729

    start(port, root, autoraise)



# -*- coding: utf-8 -*-

"""livereload.task

Task management for LiveReload Server.

A basic syntax overview::

    from livereload.task import Task

    Task.add('file.css')

    def do_some_thing():
        pass

    Task.add('file.css', do_some_thing)
"""

import os
import glob
import logging

try:
    import pyinotify
    from tornado import ioloop

    class TaskEventHandler(pyinotify.ProcessEvent):
        def my_init(self, **kwargs):
            self.func = kwargs['func']

        def process_default(self, event):
            if Task.watch():
                self.func()

    HAS_PYINOTIFY = True
except ImportError:
    HAS_PYINOTIFY = False

IGNORE = [
    '.pyc', '.pyo', '.o', '.swp'
]


class Task(object):
    tasks = {}
    _modified_times = {}
    last_modified = None
    if HAS_PYINOTIFY:
        wm = pyinotify.WatchManager()
        notifier = None

    @classmethod
    def add(cls, path, func=None):
        logging.info('Add task: %s' % path)
        if HAS_PYINOTIFY:
            cls.wm.add_watch(path, pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_MODIFY, rec=True, do_glob=True, auto_add=True)
        cls.tasks[path] = func

    @classmethod
    def start(cls, func):
        if HAS_PYINOTIFY:
            if not cls.notifier:
                cls.notifier = pyinotify.TornadoAsyncNotifier(cls.wm, ioloop.IOLoop.instance(), default_proc_fun=TaskEventHandler(func=func))
                Task.watch()  # initial run so we don't miss the first change
        return HAS_PYINOTIFY

    @classmethod
    def watch(cls):
        _changed = False
        for path in cls.tasks:
            if cls.is_changed(path):
                _changed = True
                func = cls.tasks[path]
                func and func()

        return _changed

    @classmethod
    def is_changed(cls, path):
        def is_file_changed(path):
            if not os.path.isfile(path):
                return False

            _, ext = os.path.splitext(path)
            if ext in IGNORE:
                return False

            modified = int(os.stat(path).st_mtime)

            if path not in cls._modified_times:
                cls._modified_times[path] = modified
                return True

            if path in cls._modified_times and \
               cls._modified_times[path] != modified:
                logging.info('file changed: %s' % path)
                cls._modified_times[path] = modified
                cls.last_modified = path
                return True

            cls._modified_times[path] = modified
            return False

        def is_folder_changed(path):
            _changed = False
            for root, dirs, files in os.walk(path, followlinks=True):
                if '.git' in dirs:
                    dirs.remove('.git')
                if '.hg' in dirs:
                    dirs.remove('.hg')
                if '.svn' in dirs:
                    dirs.remove('.svn')
                if '.cvs' in dirs:
                    dirs.remove('.cvs')

                for f in files:
                    if is_file_changed(os.path.join(root, f)):
                        _changed = True

            return _changed

        def is_glob_changed(path):
            _changed = False
            for f in glob.glob(path):
                if is_file_changed(f):
                    _changed = True

            return _changed

        if os.path.isfile(path):
            return is_file_changed(path)
        elif os.path.isdir(path):
            return is_folder_changed(path)
        else:
            return is_glob_changed(path)
        return False



# -*- coding: utf-8 -*-

"""livereload.app

Core Server of LiveReload.
"""

import os
import logging
import time
import mimetypes
import webbrowser
import hashlib
from tornado import ioloop
from tornado import escape
from tornado import websocket
from tornado.web import RequestHandler, Application
from tornado.util import ObjectDict
try:
    from tornado.log import enable_pretty_logging
except ImportError:
    from tornado.options import enable_pretty_logging
from livereload.task import Task


PORT = 35729
ROOT = '.'
LIVERELOAD = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    'livereload.js',
)


class LiveReloadHandler(websocket.WebSocketHandler):
    waiters = set()
    _last_reload_time = None

    def allow_draft76(self):
        return True

    def on_close(self):
        if self in LiveReloadHandler.waiters:
            LiveReloadHandler.waiters.remove(self)

    def send_message(self, message):
        if isinstance(message, dict):
            message = escape.json_encode(message)

        try:
            self.write_message(message)
        except:
            logging.error('Error sending message', exc_info=True)

    def poll_tasks(self):
        changes = Task.watch()
        if not changes:
            return
        self.watch_tasks()

    def watch_tasks(self):
        if time.time() - self._last_reload_time < 3:
            # if you changed lot of files in one time
            # it will refresh too many times
            logging.info('ignore this reload action')
            return

        logging.info('Reload %s waiters', len(self.waiters))

        msg = {
            'command': 'reload',
            'path': Task.last_modified or '*',
            'liveCSS': True
        }

        self._last_reload_time = time.time()
        for waiter in LiveReloadHandler.waiters:
            try:
                waiter.write_message(msg)
            except:
                logging.error('Error sending message', exc_info=True)
                LiveReloadHandler.waiters.remove(waiter)

    def on_message(self, message):
        """Handshake with livereload.js

        1. client send 'hello'
        2. server reply 'hello'
        3. client send 'info'

        http://help.livereload.com/kb/ecosystem/livereload-protocol
        """
        message = ObjectDict(escape.json_decode(message))
        if message.command == 'hello':
            handshake = {}
            handshake['command'] = 'hello'
            protocols = message.protocols
            protocols.append(
                'http://livereload.com/protocols/2.x-remote-control'
            )
            handshake['protocols'] = protocols
            handshake['serverName'] = 'livereload-tornado'
            self.send_message(handshake)

        if message.command == 'info' and 'url' in message:
            logging.info('Browser Connected: %s' % message.url)
            LiveReloadHandler.waiters.add(self)
            if not LiveReloadHandler._last_reload_time:
                if os.path.exists('Guardfile'):
                    logging.info('Reading Guardfile')
                    execfile('Guardfile', {})
                else:
                    logging.info('No Guardfile')
                    Task.add(os.getcwd())

                LiveReloadHandler._last_reload_time = time.time()
                logging.info('Start watching changes')
                if not Task.start(self.watch_tasks):
                    ioloop.PeriodicCallback(self.poll_tasks, 800).start()


class IndexHandler(RequestHandler):

    def get(self, path='/'):
        abspath = os.path.join(os.path.abspath(ROOT), path.lstrip('/'))
        mime_type, encoding = mimetypes.guess_type(abspath)
        if not mime_type:
            mime_type = 'text/html'

        self.mime_type = mime_type
        self.set_header('Content-Type', mime_type)
        self.read_path(abspath)

    def inject_livereload(self):
        if self.mime_type != 'text/html':
            return
        ua = self.request.headers.get('User-Agent', 'bot').lower()
        if 'msie' not in ua:
            self.write('<script src="/livereload.js"></script>')

    def read_path(self, abspath):
        filepath = abspath
        if os.path.isdir(filepath):
            filepath = os.path.join(abspath, 'index.html')
            if not os.path.exists(filepath):
                self.create_index(abspath)
                return
        elif not os.path.exists(abspath):
            filepath = abspath + '.html'

        if os.path.exists(filepath):
            if self.mime_type == 'text/html':
                f = open(filepath)
                data = f.read()
                f.close()
                before, after = data.split('</head>')
                self.write(before)
                self.inject_livereload()
                self.write('</head>')
                self.write(after)
            else:
                f = open(filepath, 'rb')
                data = f.read()
                f.close()
                self.write(data)

            hasher = hashlib.sha1()
            hasher.update(data)
            self.set_header('Etag', '"%s"' % hasher.hexdigest())
            return
        self.send_error(404)
        return

    def create_index(self, root):
        self.inject_livereload()
        files = os.listdir(root)
        self.write('<ul>')
        for f in files:
            path = os.path.join(root, f)
            self.write('<li>')
            if os.path.isdir(path):
                self.write('<a href="%s/">%s</a>' % (f, f))
            else:
                self.write('<a href="%s">%s</a>' % (f, f))
            self.write('</li>')

        self.write('</ul>')


class LiveReloadJSHandler(RequestHandler):
    def get(self):
        f = open(LIVERELOAD)
        self.set_header('Content-Type', 'application/javascript')
        for line in f:
            if '{{port}}' in line:
                line = line.replace('{{port}}', str(PORT))
            self.write(line)
        f.close()

handlers = [
    (r'/livereload', LiveReloadHandler),
    (r'/livereload.js', LiveReloadJSHandler),
    (r'(.*)', IndexHandler),
]


def start(port=35729, root='.', autoraise=False):
    global PORT
    PORT = port
    global ROOT
    if root is None:
        root = '.'
    ROOT = root
    logging.getLogger().setLevel(logging.INFO)
    enable_pretty_logging()
    app = Application(handlers=handlers)
    app.listen(port)
    print('Serving path %s on 127.0.0.1:%s' % (root, port))

    if autoraise:
        webbrowser.open(
            'http://127.0.0.1:%s' % port, new=2, autoraise=True
        )
    try:
        ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        print('Shutting down...')


if __name__ == '__main__':
    start(8000)



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""livereload.compiler

Provides a set of compilers for web developers.

Available compilers now:

+ less
+ coffee
+ uglifyjs
+ slimmer
"""

import os
import functools
import logging
from subprocess import Popen, PIPE


def make_folder(dest):
    folder = os.path.split(dest)[0]
    if not folder:
        return
    if os.path.isdir(folder):
        return
    try:
        os.makedirs(folder)
    except:
        pass


def _get_http_file(url, build_dir='build/assets'):
    import hashlib
    key = hashlib.md5(url).hexdigest()
    filename = os.path.join(os.getcwd(), build_dir, key)
    if os.path.exists(filename):
        return filename
    make_folder(filename)

    import urllib
    print('Downloading: %s' % url)
    urllib.urlretrieve(url, filename)
    return filename


class BaseCompiler(object):
    """BaseCompiler

    BaseCompiler defines the basic syntax of a Compiler.

    >>> c = BaseCompiler('a')
    >>> c.write('b')  #: write compiled code to 'b'
    >>> c.append('c')  #: append compiled code to 'c'
    """
    def __init__(self, path=None):
        if path:
            if path.startswith('http://') or path.startswith('https://'):
                path = _get_http_file(path)
            self.filetype = os.path.splitext(path)[1]
        self.path = path

    def get_code(self):
        f = open(self.path)
        code = f.read()
        f.close()
        return code

    def write(self, output):
        """write code to output"""
        logging.info('write %s' % output)
        make_folder(output)
        f = open(output, 'w')
        code = self.get_code()
        if code:
            f.write(code)
        f.close()

    def append(self, output):
        """append code to output"""
        logging.info('append %s' % output)
        make_folder(output)
        f = open(output, 'a')
        f.write(self.get_code())
        f.close()

    def __call__(self, output, mode='w'):
        if mode == 'a':
            self.append(output)
            return
        self.write(output)
        return


class CommandCompiler(BaseCompiler):
    def init_command(self, command, source=None):
        self.command = command
        self.source = source

    def get_code(self):
        cmd = self.command.split()
        if self.path:
            cmd.append(self.path)

        try:
            p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        except OSError as e:
            logging.error(e)
            if e.errno == os.errno.ENOENT:  # file (command) not found
                logging.error("maybe you haven't installed %s", cmd[0])
            return None
        if self.source:
            stdout, stderr = p.communicate(input=self.source)
        else:
            stdout, stderr = p.communicate()
        if stderr:
            logging.error(stderr)
            return None
        #: stdout is bytes, decode for python3
        return stdout.decode()


def lessc(path, output, mode='w'):
    _compile = CommandCompiler(path)
    _compile.init_command('lessc --compress')
    return functools.partial(_compile, output, mode)


def uglifyjs(path, output, mode='w'):
    _compile = CommandCompiler(path)
    _compile.init_command('uglifyjs --nc')
    return functools.partial(_compile, output, mode)


class SlimmerCompiler(BaseCompiler):
    def get_code(self):
        import slimmer
        f = open(self.path)
        code = f.read()
        f.close()
        if self.filetype == '.css':
            return slimmer.css_slimmer(code)
        if self.filetype == '.js':
            return slimmer.js_slimmer(code)
        if self.filetype == '.html':
            return slimmer.xhtml_slimmer(code)
        return code


def slimmer(path, output, mode='w'):
    _compile = SlimmerCompiler(path)
    return functools.partial(_compile, output, mode)


def rstc(path, output, mode='w'):
    _compile = CommandCompiler(path)
    _compile.init_command('rst2html.py')
    return functools.partial(_compile, output, mode)


def shell(command, path=None, output=os.devnull, mode='w'):
    _compile = CommandCompiler(path)
    _compile.init_command(command)
    return functools.partial(_compile, output, mode)


def coffee(path, output, mode='w'):
    _compile = CommandCompiler(path)
    f = open(path)
    code = f.read()
    f.close()
    _compile.init_command('coffee --compile --stdio', code)
    return functools.partial(_compile, output, mode)
